package eu.eudat.integrationevent.outbox.notification;

import eu.eudat.audit.AuditableAction;
import eu.eudat.commons.enums.ContactInfoType;
import eu.eudat.commons.enums.notification.NotificationContactType;
import eu.eudat.commons.enums.notification.NotificationNotifyState;
import eu.eudat.commons.enums.notification.NotificationTrackingProcess;
import eu.eudat.commons.enums.notification.NotificationTrackingState;
import eu.eudat.commons.validation.ValidatorFactory;
import eu.eudat.data.UserContactInfoEntity;
import eu.eudat.data.UserEntity;
import eu.eudat.integrationevent.outbox.OutboxService;
import eu.eudat.model.persist.notification.NotificationPersist;
import eu.eudat.query.UserContactInfoQuery;
import eu.eudat.query.UserQuery;
import eu.eudat.service.notification.NotificationService;
import gr.cite.tools.auditing.AuditService;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.logging.LoggerService;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.web.context.annotation.RequestScope;

import javax.management.InvalidApplicationException;
import java.time.Instant;
import java.util.List;

@Component
@RequestScope
public class NotificationIntegrationEventHandlerImpl implements NotificationIntegrationEventHandler {

    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(NotificationIntegrationEventHandlerImpl.class));

    private final NotificationService notificationService;

    private final QueryFactory queryFactory;

    private final AuditService auditService;

    private final OutboxService outboxService;

    private final ValidatorFactory validatorFactory;

    @Autowired
    public NotificationIntegrationEventHandlerImpl(
            OutboxService outboxService,
            NotificationService notificationService,
            QueryFactory queryFactory,
            AuditService auditService, ValidatorFactory validatorFactory) {
        this.outboxService = outboxService;
        this.notificationService = notificationService;
        this.validatorFactory = validatorFactory;
        this.queryFactory = queryFactory;
        this.auditService = auditService;
    }

    @Override
    public void handle(NotificationIntegrationEvent event) throws InvalidApplicationException {
//		OutboxIntegrationEvent message = new OutboxIntegrationEvent();
//		message.setMessageId(UUID.randomUUID());
//		message.setType(OutboxIntegrationEvent.NOTIFY);
//		message.setEvent(event);
//		this.outboxService.publish(message);
        NotificationPersist persist = new NotificationPersist();
        persist.setType(event.getNotificationType());
        persist.setUserId(event.getUserId());
        persist.setContactHint(event.getContactHint());
        persist.setContactTypeHint(event.getContactTypeHint());
        persist.setData(event.getData());
        persist.setNotifyState(NotificationNotifyState.PENDING);
        persist.setNotifiedWith(NotificationContactType.EMAIL);
        persist.setRetryCount(0);
        persist.setTrackingState(NotificationTrackingState.UNDEFINED);
        persist.setTrackingProcess(NotificationTrackingProcess.PENDING);
        persist.setTrackingData(null);
        persist.setProvenanceRef(event.getProvenanceRef());
        persist.setNotifiedAt(Instant.now());
        validatorFactory.validator(NotificationPersist.NotificationPersistValidator.class).validateForce(persist);
        if (isNotificationConsistent(persist)) {
            notificationService.persist(persist, null);
            auditService.track(AuditableAction.Notification_Persist, "notification_event", event);
        }
    }

    private boolean isNotificationConsistent(NotificationPersist notification) {
        switch (notification.getContactTypeHint()) {
            case IN_APP: {
                if (notification.getUserId() == null)
                    return false;

                List<UserEntity> users = this.queryFactory.query(UserQuery.class).ids(notification.getUserId()).collect();
                return !users.isEmpty();
            }
            case EMAIL: {
                if (notification.getContactHint() != null && !notification.getContactHint().isBlank())
                    return true;

                if (notification.getUserId() == null)
                    return false;

                List<UserContactInfoEntity> userContactInfoEntities = this.queryFactory.query(UserContactInfoQuery.class).types(ContactInfoType.Email).userIds(notification.getUserId()).collect();
                return !userContactInfoEntities.isEmpty();
            }
            default:
                throw new MyApplicationException("invalid type " + notification.getContactTypeHint());

        }
    }
}
