package eu.eudat.model.builder.deposit;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.commons.enums.IsActive;
import eu.eudat.convention.ConventionService;
import eu.eudat.data.*;
import eu.eudat.depositinterface.enums.DmpAccessType;
import eu.eudat.depositinterface.models.*;
import eu.eudat.model.*;
import eu.eudat.query.*;
import gr.cite.tools.data.builder.BuilderFactory;
import gr.cite.tools.data.query.Ordering;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.fieldset.BaseFieldSet;
import gr.cite.tools.logging.LoggerService;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.*;
import java.util.stream.Collectors;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class DmpDepositBuilder extends BaseDepositBuilder<DmpDepositModel, DmpEntity> {

    private final QueryFactory queryFactory;

    private final BuilderFactory builderFactory;
    private FileEnvelope pdfFile;
    private FileEnvelope rdaJsonFile;
    private FileEnvelope supportingFilesZip;
    private String repositoryId;
    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);

    @Autowired
    public DmpDepositBuilder(ConventionService conventionService,
                             QueryFactory queryFactory,
                             BuilderFactory builderFactory) {
        super(conventionService, new LoggerService(LoggerFactory.getLogger(DmpDepositBuilder.class)));
        this.queryFactory = queryFactory;
        this.builderFactory = builderFactory;
    }

    public DmpDepositBuilder authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    public DmpDepositBuilder setPdfFile(FileEnvelope pdfFile) {
        this.pdfFile = pdfFile;
        return this;
    }

    public DmpDepositBuilder setRdaJsonFile(FileEnvelope rdaJsonFile) {
        this.rdaJsonFile = rdaJsonFile;
        return this;
    }

    public DmpDepositBuilder setSupportingFilesZip(FileEnvelope supportingFilesZip) {
        this.supportingFilesZip = supportingFilesZip;
        return this;
    }

    public DmpDepositBuilder setRepositoryId(String repositoryId) {
        this.repositoryId = repositoryId;
        return this;
    }

    @Override
    protected List<DepositBuilderItemResponse<DmpDepositModel, DmpEntity>> buildInternal(List<DmpEntity> data) throws MyApplicationException {
        this.logger.debug("building for {}", Optional.ofNullable(data).map(List::size).orElse(0));
        if (data == null || data.isEmpty()) return new ArrayList<>();
        
        List<DepositBuilderItemResponse<DmpDepositModel, DmpEntity>> models = new ArrayList<>();

        Map<UUID, List<ReferenceDepositModel>> dmpReferencesMap = this.collectReferences(data);

        Map<UUID, List<DmpUserDepositModel>> dmpUsersMap = this.collectDmpUsers(data);

        Map<UUID, List<DescriptionDepositModel>> descriptionsMap = this.collectDmpDescriptions(data);

        for (DmpEntity d : data) {
            DmpDepositModel m = new DmpDepositModel();
            m.setId(d.getId());
            m.setLabel(d.getLabel());
            m.setVersion(d.getVersion());
            m.setDescription(d.getDescription());
            m.setFinalizedAt(d.getFinalizedAt());
            m.setPdfFile(this.pdfFile);
            m.setRdaJsonFile(this.rdaJsonFile);
            m.setSupportingFilesZip(this.supportingFilesZip);
            if (d.getVersion() > (short)1) m.setPreviousDOI(this.getPreviousDOI(d.getGroupId(), d.getId()));
            switch (d.getAccessType()){
                case Public -> m.setAccessType(DmpAccessType.Public);
                case Restricted -> m.setAccessType(DmpAccessType.Restricted);
                case null -> m.setAccessType(null);
                default -> throw new MyApplicationException("unrecognized type " + d.getAccessType().getValue());
            }
            if (dmpReferencesMap != null && !dmpReferencesMap.isEmpty() && dmpReferencesMap.containsKey(d.getId())) m.setReferences(dmpReferencesMap.get(d.getId()));
            if (dmpUsersMap != null && !dmpUsersMap.isEmpty() && dmpUsersMap.containsKey(d.getId())) m.setUsers(dmpUsersMap.get(d.getId()));
            if (descriptionsMap != null && !descriptionsMap.isEmpty() && descriptionsMap.containsKey(d.getId())) m.setDescriptions(descriptionsMap.get(d.getId()));

            models.add(new DepositBuilderItemResponse<>(m, d));
        }
        this.logger.debug("build {} items", Optional.of(models).map(List::size).orElse(0));

        return models;
    }

    private String getPreviousDOI(UUID dmpGroup, UUID currentDmpId) {
        if (repositoryId == null || repositoryId.isEmpty()) throw new MyApplicationException("repositoryId required");

        //GK: Step one get the previous version of the Data management plan
        DmpQuery dmpQuery = this.queryFactory.query(DmpQuery.class);
        dmpQuery.setOrder(new Ordering().addDescending(Dmp._version));
        List<UUID> dmpIds = dmpQuery.groupIds(dmpGroup).excludedIds(currentDmpId).isActive(IsActive.Active).collectAs(new BaseFieldSet().ensure(Dmp._id)).stream().map(DmpEntity::getId).toList();

        //GK: Step two get it's doiEntity
        List<EntityDoiEntity> dois = this.queryFactory.query(EntityDoiQuery.class).repositoryIds(repositoryId).entityIds(dmpIds).isActive(IsActive.Active).collectAs(new BaseFieldSet().ensure(EntityDoi._entityId, EntityDoi._doi));
        for(UUID uuid: dmpIds) {
            EntityDoiEntity doiEntity = dois.stream().filter(x -> x.getEntityId().equals(uuid)).findFirst().orElse(null);
            if (doiEntity != null) return doiEntity.getDoi();
        }
        return null;
    }

    private Map<UUID, List<ReferenceDepositModel>> collectReferences(List<DmpEntity> data) throws MyApplicationException {
        if (data.isEmpty()) return null;
        this.logger.debug("checking related - {}", DmpReference.class.getSimpleName());

        List<DmpReferenceEntity> dmpReferences = this.queryFactory.query(DmpReferenceQuery.class).isActives(IsActive.Active).authorize(this.authorize).dmpIds(data.stream().map(DmpEntity::getId).distinct().collect(Collectors.toList())).collectAs(new BaseFieldSet().ensure(DmpReference._dmp).ensure(DmpReference._reference));
        
        Map<UUID, List<ReferenceDepositModel>> itemMap = new HashMap<>();
        ReferenceQuery query = this.queryFactory.query(ReferenceQuery.class).authorize(this.authorize).isActive(IsActive.Active).ids(dmpReferences.stream().map(DmpReferenceEntity::getReferenceId).distinct().collect(Collectors.toList()));
        Map<UUID, ReferenceDepositModel> referenceDepositModelMap = this.builderFactory.builder(ReferenceDepositBuilder.class).authorize(this.authorize).asForeignKey(query, ReferenceEntity::getId);
        if (referenceDepositModelMap == null) return null;
        for (DmpReferenceEntity dmpReference : dmpReferences) {
            ReferenceDepositModel model  =referenceDepositModelMap.getOrDefault(dmpReference.getReferenceId(), null);
            if (model == null) continue;;
            UUID key = dmpReference.getDmpId();
            if (!itemMap.containsKey(key)) itemMap.put(key, new ArrayList<>());
            itemMap.get(key).add(model);
        }
        return itemMap;
    }


    private Map<UUID, List<DmpUserDepositModel>> collectDmpUsers(List<DmpEntity> data) throws MyApplicationException {
        this.logger.debug("checking related - {}", DmpUser.class.getSimpleName());

        Map<UUID, List<DmpUserDepositModel>> itemMap;
        DmpUserQuery query = this.queryFactory.query(DmpUserQuery.class).isActives(IsActive.Active).authorize(this.authorize).dmpIds(data.stream().map(DmpEntity::getId).distinct().collect(Collectors.toList()));
        itemMap = this.builderFactory.builder(DmpUserDepositBuilder.class).authorize(this.authorize).asMasterKey(query, DmpUserEntity::getDmpId);

        return itemMap;
    }

    private Map<UUID, List<DescriptionDepositModel>> collectDmpDescriptions(List<DmpEntity> data) throws MyApplicationException {
        if (data.isEmpty()) return null;
        this.logger.debug("checking related - {}", Description.class.getSimpleName());

        Map<UUID, List<DescriptionDepositModel>> itemMap;
        DescriptionQuery query = this.queryFactory.query(DescriptionQuery.class).isActive(IsActive.Active).authorize(this.authorize).dmpIds(data.stream().map(DmpEntity::getId).distinct().collect(Collectors.toList()));
        itemMap = this.builderFactory.builder(DescriptionDepositBuilder.class).authorize(this.authorize).asMasterKey(query, DescriptionEntity::getDmpId);

        return itemMap;
    }

}
