package eu.eudat.model.builder.deposit;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.commons.XmlHandlingService;
import eu.eudat.commons.types.reference.DefinitionEntity;
import eu.eudat.convention.ConventionService;
import eu.eudat.data.ReferenceEntity;
import eu.eudat.depositinterface.enums.ReferenceSourceType;
import eu.eudat.depositinterface.enums.ReferenceType;
import eu.eudat.depositinterface.models.ReferenceDepositModel;
import gr.cite.tools.data.builder.BuilderFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.logging.LoggerService;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.Optional;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class ReferenceDepositBuilder extends BaseDepositBuilder<ReferenceDepositModel, ReferenceEntity> {

    private final BuilderFactory builderFactory;
    private final XmlHandlingService xmlHandlingService;
    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);

    @Autowired
    public ReferenceDepositBuilder(
            ConventionService conventionService,
            BuilderFactory builderFactory, XmlHandlingService xmlHandlingService) {
        super(conventionService, new LoggerService(LoggerFactory.getLogger(ReferenceDepositBuilder.class)));
        this.builderFactory = builderFactory;
        this.xmlHandlingService = xmlHandlingService;
    }

    public ReferenceDepositBuilder authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    @Override
    protected List<DepositBuilderItemResponse<ReferenceDepositModel, ReferenceEntity>> buildInternal(List<ReferenceEntity> data) throws MyApplicationException {
        this.logger.debug("building for {}", Optional.ofNullable(data).map(List::size).orElse(0));
        if (data == null || data.isEmpty()) return new ArrayList<>();

        List<DepositBuilderItemResponse<ReferenceDepositModel, ReferenceEntity>> models = new ArrayList<>();
        for (ReferenceEntity d : data) {
            ReferenceDepositModel m = new ReferenceDepositModel();
            m.setId(d.getId());
            m.setLabel(d.getLabel());
            if (d.getDefinition() != null){
                DefinitionEntity definition = this.xmlHandlingService.fromXmlSafe(DefinitionEntity.class, d.getDefinition());
                m.setDefinition(this.builderFactory.builder(ReferenceDefinitionDepositBuilder.class).authorize(this.authorize).build(definition));
            }
            m.setReference(d.getReference());
            m.setAbbreviation(d.getAbbreviation());
            m.setDescription(d.getDescription());
            m.setSource(d.getSource());
            switch (d.getSourceType()){
                case Internal -> m.setSourceType(ReferenceSourceType.Internal);
                case External -> m.setSourceType(ReferenceSourceType.External);
                default -> throw new MyApplicationException("unrecognized type " + d.getSourceType().getValue());
            }
            switch (d.getType()){
                case Taxonomies -> m.setType(ReferenceType.Taxonomies);
                case Licenses -> m.setType(ReferenceType.Licenses);
                case Publications -> m.setType(ReferenceType.Publications);
                case Journals -> m.setType(ReferenceType.Journals);
                case PubRepositories -> m.setType(ReferenceType.PubRepositories);
                case DataRepositories -> m.setType(ReferenceType.DataRepositories);
                case Registries -> m.setType(ReferenceType.Registries);
                case Services -> m.setType(ReferenceType.Services);
                case Project -> m.setType(ReferenceType.Project);
                case Funder -> m.setType(ReferenceType.Funder);
                case Datasets -> m.setType(ReferenceType.Datasets);
                case Organizations -> m.setType(ReferenceType.Organizations);
                case Grants -> m.setType(ReferenceType.Grants);
                case Researcher -> m.setType(ReferenceType.Researcher);
                default -> throw new MyApplicationException("unrecognized type " + d.getType().getValue());
            }
            models.add(new DepositBuilderItemResponse<>(m, d));
        }
        this.logger.debug("build {} items", Optional.of(models).map(List::size).orElse(0));
        return models;
    }
}
