package eu.eudat.model.builder.referencesearch;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.commons.enums.ReferenceFieldDataType;
import eu.eudat.commons.enums.ReferenceType;
import eu.eudat.configurations.referencetype.ReferenceTypeField;
import eu.eudat.configurations.referencetype.ReferenceTypeProperties;
import eu.eudat.convention.ConventionService;
import eu.eudat.model.builder.BaseBuilder;
import eu.eudat.model.referencedefinition.Field;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.fieldset.FieldSet;
import gr.cite.tools.logging.DataLogEntry;
import gr.cite.tools.logging.LoggerService;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.*;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class FieldSearchBuilder extends BaseBuilder<Field, Map<String, String>> {

    private final ReferenceTypeProperties properties;
    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);

    @Autowired
    public FieldSearchBuilder(
            ConventionService conventionService, ReferenceTypeProperties properties) {
        super(conventionService, new LoggerService(LoggerFactory.getLogger(FieldSearchBuilder.class)));
        this.properties = properties;
    }

    public FieldSearchBuilder authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    @Override
    public List<Field> build(FieldSet fields, List< Map<String, String>> data) throws MyApplicationException {
        this.logger.debug("building for {} items requesting {} fields", Optional.ofNullable(data).map(List::size).orElse(0), Optional.ofNullable(fields).map(FieldSet::getFields).map(Set::size).orElse(0));
        this.logger.trace(new DataLogEntry("requested fields", fields));
        if (data == null)
            return new ArrayList<>();

        List<Field> models = new ArrayList<>();
        for (Map<String, String> d : data) {

            ReferenceType referenceType = ReferenceType.valueOf(d.getOrDefault("referenceType", null));
            List<ReferenceTypeField> typeFields = this.getPropertiesFields(referenceType);
            if (typeFields.isEmpty()){
                return new ArrayList<>();
            }

            for (ReferenceTypeField typeField: typeFields){
                Field m = new Field();
                m.setCode(typeField.getCode());
                m.setDataType(ReferenceFieldDataType.valueOf(typeField.getDataType()));
                m.setValue(d.getOrDefault(typeField.getCode(), null));
                models.add(m);
            }
        }
        this.logger.debug("build {} items", Optional.of(models).map(List::size).orElse(0));
        return models;
    }

    private List<ReferenceTypeField> getPropertiesFields(ReferenceType referenceType){
        switch (referenceType){
            case Taxonomies: return properties.getTaxonomy().get("fields");
            case Licenses: return properties.getLicence().get("fields");
            case Publications: return properties.getPublication().get("fields");
            case Journals: return properties.getJournal().get("fields");
            case PubRepositories: return properties.getPubRepository().get("fields");
            case DataRepositories: return properties.getDataRepository().get("fields");
            case Registries: return properties.getRegistry().get("fields");
            case Services: return properties.getService().get("fields");
            case Organizations: return properties.getOrganisation().get("fields");
            case Datasets: return properties.getDataset().get("fields");
            case Funder: return properties.getFunder().get("fields");
            case Project: return properties.getProject().get("fields");
            case Researcher: return properties.getResearcher().get("fields");
            default: return null;
        }
    }
}
