package eu.eudat.model.builder.referencesearch;

import eu.eudat.authorization.AuthorizationFlags;
import eu.eudat.commons.XmlHandlingService;
import eu.eudat.commons.enums.ReferenceSourceType;
import eu.eudat.commons.enums.ReferenceType;
import eu.eudat.convention.ConventionService;
import eu.eudat.model.Reference;
import eu.eudat.model.builder.BaseBuilder;
import eu.eudat.model.referencedefinition.Definition;
import eu.eudat.model.referencedefinition.Field;
import gr.cite.tools.data.builder.BuilderFactory;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.fieldset.FieldSet;
import gr.cite.tools.logging.DataLogEntry;
import gr.cite.tools.logging.LoggerService;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.ConfigurableBeanFactory;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.*;

@Component
@Scope(value = ConfigurableBeanFactory.SCOPE_PROTOTYPE)
public class ReferenceSearchBuilder extends BaseBuilder<Reference, Map<String, String>> {

    private final BuilderFactory builderFactory;
    private final QueryFactory queryFactory;
    private final XmlHandlingService xmlHandlingService;
    private EnumSet<AuthorizationFlags> authorize = EnumSet.of(AuthorizationFlags.None);

    @Autowired
    public ReferenceSearchBuilder(
            ConventionService conventionService,
            BuilderFactory builderFactory, QueryFactory queryFactory, XmlHandlingService xmlHandlingService) {
        super(conventionService, new LoggerService(LoggerFactory.getLogger(ReferenceSearchBuilder.class)));
        this.builderFactory = builderFactory;
        this.queryFactory = queryFactory;
        this.xmlHandlingService = xmlHandlingService;
    }

    public ReferenceSearchBuilder authorize(EnumSet<AuthorizationFlags> values) {
        this.authorize = values;
        return this;
    }

    @Override
    public List<Reference> build(FieldSet fields, List<Map<String, String>> data) throws MyApplicationException {
        this.logger.debug("building for {} items requesting {} fields", Optional.ofNullable(data).map(List::size).orElse(0), Optional.ofNullable(fields).map(FieldSet::getFields).map(Set::size).orElse(0));
        this.logger.trace(new DataLogEntry("requested fields", fields));
        if (data == null)
            return new ArrayList<>();

        List<Reference> models = new ArrayList<>();
        for (Map<String, String> d : data) {
            Reference m = new Reference();
            if (d.containsKey("id")) m.setId(UUID.fromString(d.getOrDefault("id", null)));
            m.setLabel(d.getOrDefault("name", null));
            m.setSourceType(ReferenceSourceType.External);
            String type = d.getOrDefault("referenceType", null);
            if (type != null) m.setType(ReferenceType.valueOf(type));

            Definition definition = this.builderFactory.builder(DefinitionSearchBuilder.class).authorize(this.authorize).build(null, d);
            m.setDefinition(definition);

            switch (m.getType()){
                case Researcher:
                case Organizations:
                case Funder:
                case Grants:
                case DataRepositories:
                case PubRepositories:
                case Journals:
                case Datasets:
                case Registries:
                case Services:{
                    for (Field field : definition.getFields()) {
                        if (field.getCode().equals("pid") && field.getValue() != null) {
                            if (d.containsKey("key")) {
                                m.setReference(d.get("key") + ":" + field.getValue());
                            }
                        }
                        if (field.getCode().equals("tag") && field.getValue() != null){
                            m.setSource(field.getValue());
                        }
                    }
                }
            }

            m.setDescription(d.getOrDefault("description", null));
            m.setAbbreviation(d.getOrDefault("abbreviation", null));

            models.add(m);

        }
        this.logger.debug("build {} items", Optional.of(models).map(List::size).orElse(0));
        return models;
    }

}
