package eu.eudat.data.dao.entities;

import eu.eudat.commons.enums.DmpStatus;
import eu.eudat.commons.enums.IsActive;
import eu.eudat.data.DmpEntity;
import eu.eudat.data.dao.DatabaseAccess;
import eu.eudat.data.dao.criteria.DataManagementPlanCriteria;
import eu.eudat.data.dao.criteria.DatasetWizardUserDmpCriteria;
import eu.eudat.data.dao.databaselayer.service.DatabaseService;
import eu.eudat.data.UserEntity;
import eu.eudat.queryable.QueryableList;
import eu.eudat.queryable.types.FieldSelectionType;
import eu.eudat.queryable.types.SelectionField;
import eu.eudat.types.grant.GrantStateType;
import jakarta.persistence.criteria.Join;
import jakarta.persistence.criteria.JoinType;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

import javax.management.InvalidApplicationException;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.UUID;
import java.util.concurrent.CompletableFuture;

@Component("dMPDao")
public class DMPDaoImpl extends DatabaseAccess<DmpEntity> implements DMPDao {

    @Autowired
    public DMPDaoImpl(DatabaseService<DmpEntity> databaseService) {
        super(databaseService);
    }

	@Override
	public QueryableList<DmpEntity> getWithCriteria(DataManagementPlanCriteria criteria) {
		QueryableList<DmpEntity> query = getDatabaseService().getQueryable(DmpEntity.class);
		if (criteria.getLike() != null && !criteria.getLike().isEmpty())
			query.where((builder, root) -> builder.or(
					builder.like(builder.upper(root.get("label")), "%" + criteria.getLike().toUpperCase() + "%"),
					builder.like(builder.upper(root.get("description")), "%" + criteria.getLike().toUpperCase() + "%")));
		if (criteria.getPeriodEnd() != null)
			query.where((builder, root) -> builder.lessThan(root.get("created"), criteria.getPeriodEnd()));
		if (criteria.getPeriodStart() != null)
			query.where((builder, root) -> builder.greaterThan(root.get("created"), criteria.getPeriodStart()));
		if (criteria.getProfile() != null)
			query.where((builder, root) -> builder.equal(root.get("profile"), criteria.getProfile()));
//		if (criteria.getGrants() != null && !criteria.getGrants().isEmpty()) //TODO
//			query.where(((builder, root) -> root.get("grant").in(criteria.getGrants())));
		if (!criteria.getAllVersions())
			query.initSubQuery(String.class).where((builder, root) -> builder.equal(root.get("version"),
					query.subQueryMax((builder1, externalRoot, nestedRoot) -> builder1.and(
							builder1.equal(externalRoot.get("groupId"), nestedRoot.get("groupId")),
							builder1.notEqual(nestedRoot.get("isActive"), IsActive.Inactive.getValue())), Arrays.asList(new SelectionField(FieldSelectionType.FIELD, "version")), Integer.class)));
		if (criteria.getGroupIds() != null && !criteria.getGroupIds().isEmpty())
			query.where((builder, root) -> root.get("groupId").in(criteria.getGroupIds()));
		if (criteria.getStatus() != null) {
			if (criteria.getStatus() == DmpStatus.Finalized.getValue().intValue()) {
				query.where((builder, root) -> builder.equal(root.get("status"), DmpStatus.Finalized.getValue()));
			} else if (criteria.getStatus() == DmpStatus.Draft.getValue().intValue()) {
				query.where((builder, root) -> builder.equal(root.get("status"), DmpStatus.Draft.getValue()));
			}
		}
		if (criteria.getIsPublic()) {
			query.where(((builder, root) -> builder.equal(root.get("isPublic"), criteria.getIsPublic())));
		}
        /*if (criteria.getRole() != null) {
            if (criteria.getRole().equals(UserDMP.UserDMPRoles.OWNER.getValue())) {
                query.where((builder, root) -> builder.equal(root.join("users", JoinType.LEFT).get("role"), UserDMP.UserDMPRoles.OWNER.getValue()));
            } else if (criteria.getRole().equals(UserDMP.UserDMPRoles.USER.getValue())) {
                query.where((builder, root) -> builder.equal(root.join("users", JoinType.LEFT).get("role"), UserDMP.UserDMPRoles.USER.getValue()));
            }
        }*/
        if (criteria.getOrganisations() != null && !criteria.getOrganisations().isEmpty()) {
            query.where((builder, root) -> root.join("organisations").get("reference").in(criteria.getOrganisations()));
        }
        if (criteria.getCollaborators() != null && !criteria.getCollaborators().isEmpty()) {
            query.where((builder, root) -> root.join("users", JoinType.LEFT).join("user", JoinType.LEFT).get("id").in(criteria.getCollaborators()));
        }
        if (criteria.getDatasetTemplates() != null && !criteria.getDatasetTemplates().isEmpty()) {
            query.where((builder, root) -> root.join("associatedDmps", JoinType.LEFT).get("id").in(criteria.getDatasetTemplates()));
        }
        if (criteria.getGrantStatus() != null) {
            if (criteria.getGrantStatus().equals(GrantStateType.FINISHED.getValue().shortValue()))
                query.where((builder, root) -> builder.lessThan(root.get("grant").get("enddate"), new Date()));
            if (criteria.getGrantStatus().equals(GrantStateType.ONGOING.getValue().shortValue()))
                query.where((builder, root) ->
                        builder.or(builder.greaterThan(root.get("grant").get("enddate"), new Date())
                                , builder.isNull(root.get("grant").get("enddate"))));
        }

        if (criteria.hasDoi()) {
            query.where((builder, root) -> builder.not(builder.isNull(root.join("dois").get("id"))));
        }
        query.where((builder, root) -> builder.notEqual(root.get("isActive"), IsActive.Inactive));
        return query;
    }

    public QueryableList<DmpEntity> getAuthenticated(QueryableList<DmpEntity> query, UUID principal, List<Integer> roles) {
        if (roles != null && !roles.isEmpty()) {
            query.where((builder, root) -> {
                Join userJoin = root.join("users", JoinType.LEFT);
                return builder.and(builder.equal(userJoin.join("user", JoinType.LEFT).get("id"), principal), userJoin.get("role").in(roles));
            });
        } else {
            query.where((builder, root) -> builder.equal(root.join("users", JoinType.LEFT).join("user", JoinType.LEFT).get("id"), principal));
        }

        return query;
    }

    @Override
    @Transactional
    public DmpEntity createOrUpdate(DmpEntity item) {
        return this.getDatabaseService().createOrUpdate(item, DmpEntity.class);
    }

    @Override
    public DmpEntity find(UUID id) throws InvalidApplicationException {
        return getDatabaseService().getQueryable(DmpEntity.class).where((builder, root) -> builder.equal((root.get("id")), id)).getSingle();
    }

    @Override
    public QueryableList<DmpEntity> getUserDmps(DatasetWizardUserDmpCriteria datasetWizardUserDmpCriteria, UserEntity userInfo) {
        QueryableList<DmpEntity> query = getDatabaseService().getQueryable(DmpEntity.class).where((builder, root) -> builder.or(builder.equal(root.get("creator"), userInfo), builder.isMember(userInfo, root.get("users"))));
        if (datasetWizardUserDmpCriteria.getLike() != null && !datasetWizardUserDmpCriteria.getLike().isEmpty()) {
            query.where((builder, root) -> builder.like(root.get("label"), "%" + datasetWizardUserDmpCriteria.getLike() + "%"));
        }
        return query;
    }

    @Override
    public void delete(DmpEntity item) {
        this.getDatabaseService().delete(item);
    }

    @Override
    public QueryableList<DmpEntity> asQueryable() {
        return this.getDatabaseService().getQueryable(DmpEntity.class);
    }

    @Async
    @Override
    public CompletableFuture<DmpEntity> createOrUpdateAsync(DmpEntity item) {
        return CompletableFuture.supplyAsync(() -> this.createOrUpdate(item));
    }

    @Override
    public DmpEntity find(UUID id, String hint) {
        throw new UnsupportedOperationException();
    }
}
