package eu.old.eudat.migration;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.type.MapType;
import com.fasterxml.jackson.databind.type.TypeFactory;
import eu.eudat.commons.JsonHandlingService;
import eu.eudat.commons.XmlHandlingService;
import eu.eudat.commons.enums.*;
import eu.eudat.commons.types.dmp.DmpBlueprintValueEntity;
import eu.eudat.commons.types.dmp.DmpContactEntity;
import eu.eudat.commons.types.dmp.DmpPropertiesEntity;
import eu.eudat.commons.types.reference.DefinitionEntity;
import eu.eudat.commons.types.reference.FieldEntity;
import eu.eudat.convention.ConventionService;
import eu.eudat.data.DmpEntity;
import eu.eudat.data.DmpReferenceEntity;
import eu.eudat.data.ReferenceEntity;
import eu.eudat.query.ReferenceQuery;
import eu.old.eudat.data.dao.entities.DMPDao;
import eu.old.eudat.data.entities.DMP;
import eu.old.eudat.logic.services.operations.DatabaseRepository;
import eu.old.eudat.models.data.dmp.DataManagementPlan;
import gr.cite.tools.data.query.QueryFactory;
import gr.cite.tools.exception.MyApplicationException;
import gr.cite.tools.logging.LoggerService;
import jakarta.persistence.EntityManager;
import org.slf4j.LoggerFactory;
import org.springframework.core.env.Environment;
import org.springframework.stereotype.Service;

import java.lang.reflect.Field;
import java.time.Instant;
import java.util.*;

@Service
public class DmpMigrationService {

    private static final LoggerService logger = new LoggerService(LoggerFactory.getLogger(DmpMigrationService.class));
    private final DatabaseRepository databaseRepository;
    private final JsonHandlingService jsonHandlingService;
    private final EntityManager entityManager;
    private final ConventionService conventionService;
    private final XmlHandlingService xmlHandlingService;
    private final Environment environment;

    private static final int PageSize = 500;
    private static final boolean TestMode = false;

    public DmpMigrationService(DatabaseRepository databaseRepository, JsonHandlingService jsonHandlingService, EntityManager entityManager, ConventionService conventionService, XmlHandlingService xmlHandlingService, Environment environment) {
        this.databaseRepository = databaseRepository;
        this.jsonHandlingService = jsonHandlingService;
        this.entityManager = entityManager;
        this.conventionService = conventionService;
        this.xmlHandlingService = xmlHandlingService;
        this.environment = environment;
    }

    public void migrate() throws JsonProcessingException, NoSuchFieldException, IllegalAccessException {
        DMPDao dmpDao = databaseRepository.getDmpDao();
        long total = dmpDao.asQueryable().count();
        logger.debug("Migrate Dmp Total : " + total);
        int page = 0;

        List<DMP> items;
        do {
            items = dmpDao.asQueryable().orderBy((builder, root) -> builder.asc(root.get("created"))).orderBy((builder, root) -> builder.asc(root.get("ID"))).skip(page * PageSize).take(PageSize).toList();
            List<DMP> finalItems = items;
            List<DMP> groupDmps = dmpDao.asQueryable().where((builder, root) -> root.get("groupId").in(finalItems.stream().map(DMP::getGroupId).distinct())).toList();
            Map<UUID, List<DMP>> groupDmpMap = new HashMap<>();
            for (DMP dmp: groupDmps) {
                if (!groupDmpMap.containsKey(dmp.getGroupId())) groupDmpMap.put(dmp.getGroupId(), new ArrayList<>());
                groupDmpMap.get(dmp.getGroupId()).add(dmp);
            }
            
            if (items != null && !items.isEmpty()) {
                logger.debug("Migrate Dmp " + page * PageSize + " of " + total);

                for (DMP item : items) {
                    //entityManager.detach(item);
                    
                    DataManagementPlan model = new DataManagementPlan();
                    model.fromDataModel(item);
                    List<DMP> itemGroupDmps = groupDmpMap.get(item.getGroupId());
                    DMP currentVersionDmp = itemGroupDmps.stream().max(Comparator.comparing(DMP::getVersion)).orElseThrow();
                    
                    DmpPropertiesEntity dmpProperties = new DmpPropertiesEntity();
                    dmpProperties.setDmpBlueprintValues(new ArrayList<>());
                    dmpProperties.setContacts(new ArrayList<>());

                    DmpEntity data = new DmpEntity();
                    data.setId(model.getId());
                    data.setLabel(model.getLabel());
                    data.setDescription(model.getDescription());
                    if (model.getCreator() != null && model.getCreator().getId() != null) {
                        data.setCreatorId(model.getCreator().getId());
                    } else {
                        for (DMP groupDmp: itemGroupDmps) {
                            DataManagementPlan groupDmpModel = new DataManagementPlan();
                            groupDmpModel.fromDataModel(groupDmp);
                            if (groupDmpModel.getCreator() != null) {
                                data.setCreatorId(groupDmpModel.getCreator().getId());
                                break;
                            }
                        }
                    }
                    data.setGroupId(model.getGroupId());
                    data.setVersion((short) model.getVersion());
                    data.setVersionStatus(currentVersionDmp.getId().equals(item.getId()) ? DmpVersionStatus.Current: DmpVersionStatus.Previous);
                    if (model.getCreated() != null)
                        data.setCreatedAt(model.getCreated().toInstant());
                    if (model.getModified() != null)
                        data.setUpdatedAt(model.getModified().toInstant());
                    if (model.getFinalized() != null)
                        data.setFinalizedAt(model.getFinalized().toInstant());
                    data.setBlueprintId(model.getProfile().getId());
                    if (model.getExtraProperties() != null) {
                        if (model.getExtraProperties().containsKey("language") && model.getExtraProperties().get("language") != null)
                            data.setLanguage((String) model.getExtraProperties().get("language"));
                        if (model.getExtraProperties().containsKey("visible") && model.getExtraProperties().get("visible") != null)
                            data.setAccessType((boolean) model.getExtraProperties().get("visible") ? DmpAccessType.Public : DmpAccessType.Restricted);
                        if (model.getExtraProperties().containsKey("contact") && model.getExtraProperties().get("contact") != null) {
                            DmpContactEntity contactEntity = new DmpContactEntity();
                            contactEntity.setUserId((String) model.getExtraProperties().get("contact"));
                            dmpProperties.getContacts().add(contactEntity);
                        }
                    }
                    if (model.getProperties() != null) {
                        model.getProperties().forEach((key,val) -> {
                            DmpBlueprintValueEntity valueEntity = new DmpBlueprintValueEntity();
                            valueEntity.setFieldId(key);
                            valueEntity.setValue((String) val);
                            dmpProperties.getDmpBlueprintValues().add(valueEntity);
                        });
                    }
                    if (model.getStatus() == 99) {
                        data.setIsActive(IsActive.Inactive);
                        data.setStatus(DmpStatus.Draft);
                    } else {
                        data.setIsActive(IsActive.Active);
                        data.setStatus(DmpStatus.of((short) model.getStatus()));
                    }
                    data.setProperties(jsonHandlingService.toJson(dmpProperties));

                    if (data.getAccessType() == null && item.getStatus().equals(DmpStatus.Finalized.getValue())) {
                        logger.warn("AccessType not found set to default for dmp " + item.getId());
                        data.setAccessType(item.isPublic() ? DmpAccessType.Public : DmpAccessType.Restricted);
                    }
                    if ((data.getLanguage() == null || data.getLanguage().isBlank()) && item.getStatus().equals(DmpStatus.Finalized.getValue())) {
                        logger.warn("Language not found set to default for dmp " + item.getId());
                        data.setLanguage("en");
                    }
                    if (data.getCreatorId() == null){
                        logger.warn("Migration skipped creator not found " + item.getId());
                        throw new MyApplicationException("Migration skipped creator not found " + item.getId());
                    }

                    this.entityManager.persist(data);
                    this.entityManager.flush();
                }

                page++;
            }
        } while (items != null && !items.isEmpty() && !TestMode);
    }

    public void migrateDmpLicenses() throws JsonProcessingException {
        DMPDao dmpDao = databaseRepository.getDmpDao();
        long total = dmpDao.asQueryable().count();
        logger.debug("Migrate Licenses for Dmp Total : " + total);
        int page = 0;

        Set<String> collectedLicenses = new HashSet<>();
        Map<String, UUID> licenseIdByName = new HashMap<>();
        List<DMP> items;
        do {
            items = dmpDao.asQueryable().orderBy((builder, root) -> builder.asc(root.get("created"))).orderBy((builder, root) -> builder.asc(root.get("ID"))).skip(page * PageSize).take(PageSize).toList();

            if (items != null && !items.isEmpty()) {
                logger.debug("Migrate Licenses for Dmp " + page * PageSize + " of " + total);

                for (DMP item : items) {
                    DataManagementPlan model = new DataManagementPlan();
                    model.fromDataModel(item);
                    if (model.getExtraProperties() != null) {
                        if (model.getExtraProperties().containsKey("license") && model.getExtraProperties().get("license") != null) {
                            Object license = model.getExtraProperties().get("license");
                            HashMap<String, String> licenseMap = this.jsonHandlingService.mapFromJson(this.jsonHandlingService.toJson(license));
                            ReferenceEntity referenceEntity = new ReferenceEntity();
                            String licensePid;
                            if (licenseMap.containsKey("pid") && licenseMap.get("pid") != null) {
                                licensePid = licenseMap.get("pid");
                                referenceEntity.setReference(licensePid);
                            }
                            else
                                continue;
                            boolean licenseExists = collectedLicenses.contains(licensePid);
                            if (!licenseExists) {
                                DefinitionEntity definitionEntity = new DefinitionEntity();
                                List<FieldEntity> fields = new ArrayList<>();
                                if (licenseMap.containsKey("name")) {
                                    referenceEntity.setLabel(licenseMap.get("name"));
                                }
                                if (licenseMap.containsKey("pid")) {
                                    if (!this.conventionService.isNullOrEmpty(licensePid)){
                                        FieldEntity fieldEntity = new FieldEntity();
                                        fieldEntity.setCode("pid");
                                        fieldEntity.setDataType(ReferenceFieldDataType.Text);
                                        fieldEntity.setValue(licensePid);
                                        fields.add(fieldEntity);
                                    }
                                }
                                if (licenseMap.containsKey("tag")) {
                                    if (!this.conventionService.isNullOrEmpty(licenseMap.get("tag"))){
                                        FieldEntity fieldEntity = new FieldEntity();
                                        fieldEntity.setCode("tag");
                                        fieldEntity.setDataType(ReferenceFieldDataType.Text);
                                        fieldEntity.setValue(licenseMap.get("tag"));
                                        fields.add(fieldEntity);
                                    }
                                }
                                if (licenseMap.containsKey("uri")) {
                                    referenceEntity.setReference(licenseMap.get("uri"));
                                    if (!this.conventionService.isNullOrEmpty(licenseMap.get("uri"))){
                                        FieldEntity fieldEntity = new FieldEntity();
                                        fieldEntity.setCode("uri");
                                        fieldEntity.setDataType(ReferenceFieldDataType.Text);
                                        fieldEntity.setValue(licenseMap.get("uri"));
                                        fields.add(fieldEntity);
                                    }
                                }
                                if (licenseMap.containsKey("abbreviation")) {
                                    referenceEntity.setAbbreviation(licenseMap.get("abbreviation"));
                                }
                                if (licenseMap.containsKey("created") && licenseMap.get("created") != null) {
                                    referenceEntity.setUpdatedAt(Instant.parse(licenseMap.get("created")));
                                } else {
                                    referenceEntity.setCreatedAt(item.getCreated() == null ? Instant.now() : item.getCreated().toInstant());
                                }
                                if (licenseMap.containsKey("modified") && licenseMap.get("modified") != null) {
                                    referenceEntity.setUpdatedAt(Instant.parse(licenseMap.get("modified")));
                                } else {
                                    referenceEntity.setUpdatedAt(item.getModified() == null ? Instant.now() : item.getModified().toInstant());
                                }
                                referenceEntity.setId(UUID.randomUUID());
                                referenceEntity.setSourceType(ReferenceSourceType.External);
                                referenceEntity.setSource(this.environment.getProperty("migration.default-license-source", "null"));
                                referenceEntity.setType(ReferenceType.Licenses);
                                referenceEntity.setIsActive(IsActive.Active);
                                definitionEntity.setFields(fields);
                                referenceEntity.setDefinition(this.xmlHandlingService.toXmlSafe(definitionEntity));

                                collectedLicenses.add(licensePid);
                                licenseIdByName.put(licensePid, referenceEntity.getId());
                                logger.debug("License '{}' migrated", licensePid);
                            }

                            DmpReferenceEntity dmpReferenceEntity = new DmpReferenceEntity();
                            dmpReferenceEntity.setId(UUID.randomUUID());
                            dmpReferenceEntity.setDmpId(item.getId());
                            dmpReferenceEntity.setReferenceId(licenseIdByName.get(licensePid));
                            dmpReferenceEntity.setCreatedAt(Instant.now());
                            dmpReferenceEntity.setUpdatedAt(Instant.now());
                            dmpReferenceEntity.setIsActive(IsActive.Active);

                            if (!licenseExists)
                                entityManager.persist(referenceEntity);
                            entityManager.persist(dmpReferenceEntity);
                        }
                    }
                    this.entityManager.flush();
                }

                page++;
            }
        } while (items != null && !items.isEmpty() && !TestMode);
        logger.info("Dmp licenses migration finished");
    }

}
